#!/usr/bin/python

#Audio Tools, a module and set of tools for manipulating audio data
#Copyright (C) 2007-2012  Brian Langenberger

#This program is free software; you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation; either version 2 of the License, or
#(at your option) any later version.

#This program is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU General Public License for more details.

#You should have received a copy of the GNU General Public License
#along with this program; if not, write to the Free Software
#Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA


def read_ogg_packets(reader):
    """given a BitstreamReader
    yields BitstreamReader substream objects
    for each packet within the Ogg stream"""

    from .bitstream import Substream

    header_type = 0
    packet = Substream(1)

    while (not (header_type & 0x4)):
        (magic_number,
         version,
         header_type,
         granule_position,
         serial_number,
         page_sequence_number,
         checksum,
         segment_count) = reader.parse("4b 8u 8u 64S 32u 32u 32u 8u")
        for segment_length in [reader.read(8) for i in xrange(segment_count)]:
            reader.substream_append(packet, segment_length)
            if (segment_length != 255):
                yield packet
                packet = Substream(1)


def read_ogg_packets_data(reader):
    """given a BitstreamReader
    yields binary strings
    for each packet within the Ogg stream"""

    header_type = 0
    packet = []

    while (not (header_type & 0x4)):
        (magic_number,
         version,
         header_type,
         granule_position,
         serial_number,
         page_sequence_number,
         checksum,
         segment_count) = reader.parse("4b 8u 8u 64S 32u 32u 32u 8u")
        for segment_length in [reader.read(8) for i in xrange(segment_count)]:
            packet.append(reader.read_bytes(segment_length))
            if (segment_length != 255):
                yield "".join(packet)
                packet = []


class OggChecksum:
    """calculates the checksum of Ogg pages
    the final checksum may be determined by int(ogg_checksum_instance)"""

    CRC_LOOKUP = (0x00000000, 0x04c11db7, 0x09823b6e, 0x0d4326d9,
                  0x130476dc, 0x17c56b6b, 0x1a864db2, 0x1e475005,
                  0x2608edb8, 0x22c9f00f, 0x2f8ad6d6, 0x2b4bcb61,
                  0x350c9b64, 0x31cd86d3, 0x3c8ea00a, 0x384fbdbd,
                  0x4c11db70, 0x48d0c6c7, 0x4593e01e, 0x4152fda9,
                  0x5f15adac, 0x5bd4b01b, 0x569796c2, 0x52568b75,
                  0x6a1936c8, 0x6ed82b7f, 0x639b0da6, 0x675a1011,
                  0x791d4014, 0x7ddc5da3, 0x709f7b7a, 0x745e66cd,
                  0x9823b6e0, 0x9ce2ab57, 0x91a18d8e, 0x95609039,
                  0x8b27c03c, 0x8fe6dd8b, 0x82a5fb52, 0x8664e6e5,
                  0xbe2b5b58, 0xbaea46ef, 0xb7a96036, 0xb3687d81,
                  0xad2f2d84, 0xa9ee3033, 0xa4ad16ea, 0xa06c0b5d,
                  0xd4326d90, 0xd0f37027, 0xddb056fe, 0xd9714b49,
                  0xc7361b4c, 0xc3f706fb, 0xceb42022, 0xca753d95,
                  0xf23a8028, 0xf6fb9d9f, 0xfbb8bb46, 0xff79a6f1,
                  0xe13ef6f4, 0xe5ffeb43, 0xe8bccd9a, 0xec7dd02d,
                  0x34867077, 0x30476dc0, 0x3d044b19, 0x39c556ae,
                  0x278206ab, 0x23431b1c, 0x2e003dc5, 0x2ac12072,
                  0x128e9dcf, 0x164f8078, 0x1b0ca6a1, 0x1fcdbb16,
                  0x018aeb13, 0x054bf6a4, 0x0808d07d, 0x0cc9cdca,
                  0x7897ab07, 0x7c56b6b0, 0x71159069, 0x75d48dde,
                  0x6b93dddb, 0x6f52c06c, 0x6211e6b5, 0x66d0fb02,
                  0x5e9f46bf, 0x5a5e5b08, 0x571d7dd1, 0x53dc6066,
                  0x4d9b3063, 0x495a2dd4, 0x44190b0d, 0x40d816ba,
                  0xaca5c697, 0xa864db20, 0xa527fdf9, 0xa1e6e04e,
                  0xbfa1b04b, 0xbb60adfc, 0xb6238b25, 0xb2e29692,
                  0x8aad2b2f, 0x8e6c3698, 0x832f1041, 0x87ee0df6,
                  0x99a95df3, 0x9d684044, 0x902b669d, 0x94ea7b2a,
                  0xe0b41de7, 0xe4750050, 0xe9362689, 0xedf73b3e,
                  0xf3b06b3b, 0xf771768c, 0xfa325055, 0xfef34de2,
                  0xc6bcf05f, 0xc27dede8, 0xcf3ecb31, 0xcbffd686,
                  0xd5b88683, 0xd1799b34, 0xdc3abded, 0xd8fba05a,
                  0x690ce0ee, 0x6dcdfd59, 0x608edb80, 0x644fc637,
                  0x7a089632, 0x7ec98b85, 0x738aad5c, 0x774bb0eb,
                  0x4f040d56, 0x4bc510e1, 0x46863638, 0x42472b8f,
                  0x5c007b8a, 0x58c1663d, 0x558240e4, 0x51435d53,
                  0x251d3b9e, 0x21dc2629, 0x2c9f00f0, 0x285e1d47,
                  0x36194d42, 0x32d850f5, 0x3f9b762c, 0x3b5a6b9b,
                  0x0315d626, 0x07d4cb91, 0x0a97ed48, 0x0e56f0ff,
                  0x1011a0fa, 0x14d0bd4d, 0x19939b94, 0x1d528623,
                  0xf12f560e, 0xf5ee4bb9, 0xf8ad6d60, 0xfc6c70d7,
                  0xe22b20d2, 0xe6ea3d65, 0xeba91bbc, 0xef68060b,
                  0xd727bbb6, 0xd3e6a601, 0xdea580d8, 0xda649d6f,
                  0xc423cd6a, 0xc0e2d0dd, 0xcda1f604, 0xc960ebb3,
                  0xbd3e8d7e, 0xb9ff90c9, 0xb4bcb610, 0xb07daba7,
                  0xae3afba2, 0xaafbe615, 0xa7b8c0cc, 0xa379dd7b,
                  0x9b3660c6, 0x9ff77d71, 0x92b45ba8, 0x9675461f,
                  0x8832161a, 0x8cf30bad, 0x81b02d74, 0x857130c3,
                  0x5d8a9099, 0x594b8d2e, 0x5408abf7, 0x50c9b640,
                  0x4e8ee645, 0x4a4ffbf2, 0x470cdd2b, 0x43cdc09c,
                  0x7b827d21, 0x7f436096, 0x7200464f, 0x76c15bf8,
                  0x68860bfd, 0x6c47164a, 0x61043093, 0x65c52d24,
                  0x119b4be9, 0x155a565e, 0x18197087, 0x1cd86d30,
                  0x029f3d35, 0x065e2082, 0x0b1d065b, 0x0fdc1bec,
                  0x3793a651, 0x3352bbe6, 0x3e119d3f, 0x3ad08088,
                  0x2497d08d, 0x2056cd3a, 0x2d15ebe3, 0x29d4f654,
                  0xc5a92679, 0xc1683bce, 0xcc2b1d17, 0xc8ea00a0,
                  0xd6ad50a5, 0xd26c4d12, 0xdf2f6bcb, 0xdbee767c,
                  0xe3a1cbc1, 0xe760d676, 0xea23f0af, 0xeee2ed18,
                  0xf0a5bd1d, 0xf464a0aa, 0xf9278673, 0xfde69bc4,
                  0x89b8fd09, 0x8d79e0be, 0x803ac667, 0x84fbdbd0,
                  0x9abc8bd5, 0x9e7d9662, 0x933eb0bb, 0x97ffad0c,
                  0xafb010b1, 0xab710d06, 0xa6322bdf, 0xa2f33668,
                  0xbcb4666d, 0xb8757bda, 0xb5365d03, 0xb1f740b4)

    def __init__(self):
        self.checksum = 0

    def reset(self):
        """clears the accumulated checksum so the object may be reused"""

        self.checksum = 0

    def update(self, byte):
        """given a byte integer, updates the running checksum"""

        self.checksum = (((self.checksum << 8) ^
                          self.CRC_LOOKUP[((self.checksum >> 24) & 0xFF) ^
                                          byte]) & 0xFFFFFFFF)

    def __int__(self):
        return self.checksum


class OggStreamReader:
    def __init__(self, reader):
        """reader is a BitstreamReader object"""

        self.reader = reader

        #try to grab a few useful bits of info
        self.reader.mark()
        (magic_number,
         version,
         self.serial_number) = self.reader.parse("4b 8u 8p 64p 32u")
        if (magic_number != "OggS"):
            from .text import ERR_OGG_INVALID_MAGIC_NUMBER
            raise ValueError(ERR_OGG_INVALID_MAGIC_NUMBER)
        elif (version != 0):
            from .text import ERR_OGG_INVALID_VERSION
            raise ValueError(ERR_OGG_INVALID_VERSION)

        self.reader.rewind()
        self.reader.unmark()
        self.checksum = OggChecksum()
        self.reader.add_callback(self.checksum.update)

    def read_page(self):
        """returns a tuple of (granule_position,
                               segments,
                               continuation,
                               first_page,
                               last_page)

        raises ValueError if the page checksum is invalid"""

        self.checksum.reset()

        #grab all the header data up to page checksum
        (magic_number,
         version,
         continuation,
         first_page,
         last_page,
         granule_position,
         serial_number,
         page_sequence_number) = self.reader.parse(
             "4b 8u 1u 1u 1u 5p 64S 32u 32u")

        #update checksum with placeholder value
        old_callback = self.reader.pop_callback()
        old_callback(0)
        old_callback(0)
        old_callback(0)
        old_callback(0)
        checksum = self.reader.read(32)
        self.reader.add_callback(old_callback)

        #grab all the segment data
        segments = self.reader.parse(
            "".join(["%db" % (segment_length) for segment_length in
                     self.reader.parse("8u" * self.reader.read(8))]))

        #verify calculated checksum against found checksum
        if (int(self.checksum) != checksum):
            from .text import ERR_OGG_CHECKSUM_MISMATCH
            raise ValueError(ERR_OGG_CHECKSUM_MISMATCH)
        else:
            return (granule_position,
                    segments,
                    continuation,
                    first_page,
                    last_page)

    def pages(self):
        """yields a tuple of (granule_position,
                              segments,
                              continuation,
                              first_page,
                              last_page)

        for each page in the stream
        raises ValueError if a page checksum is invalid"""

        while (True):
            page = self.read_page()
            yield page
            if (page[-1]):
                break


class OggStreamWriter:
    def __init__(self, writer, serial_number):
        """writer is a BitstreamWriter

        serial_number is a signed integer"""

        from .bitstream import BitstreamRecorder

        self.writer = writer
        self.serial_number = serial_number
        self.sequence_number = 0
        self.temp = BitstreamRecorder(1)
        self.checksum = OggChecksum()
        self.temp.add_callback(self.checksum.update)

    def packet_to_segments(self, packet):
        """yields a segment string per segment of the packet string

        where each segment is a string up to 255 bytes long

        the final segment will be 0 bytes if the packet
        is equally divisible by 255 bytes"""

        use_pad = len(packet) % 255 == 0

        while (len(packet) > 0):
            yield packet[0:255]
            packet = packet[255:]

        if (use_pad):
            yield ""

    def segments_to_pages(self, segments):
        """given an iterator of segment strings,

        yields a list of strings where each list is up to 255 segments
        """

        page = []
        for segment in segments:
            page.append(segment)
            if (len(page) == 255):
                yield page
                page = []

        if (len(page) > 0):
            yield page

    def write_page(self, granule_position, segments,
                   continuation, first_page, last_page):
        """granule_position is a signed long

        segments is a list of strings containing binary data

        continuation, first_page and last_page indicate this page's
        position in the Ogg stream
        """

        from .bitstream import format_size

        assert(len(segments) < 0x100)
        assert(max(map(len, segments)) < 0x100)

        self.temp.reset()
        self.checksum.reset()

        #write a checksummed header to temp
        self.temp.build("4b 8u 1u 1u 1u 5p 64S 32u 32u",
                        ("OggS", 0, continuation, first_page, last_page,
                         granule_position, self.serial_number,
                         self.sequence_number))
        self.checksum.update(0)
        self.checksum.update(0)
        self.checksum.update(0)
        self.checksum.update(0)

        #write the segment lengths and segment data to temp
        self.temp.write(8, len(segments))
        self.temp.build("8u" * len(segments), map(len, segments))
        self.temp.build("".join(["%db" % (len(segment))
                                 for segment in segments]), segments)

        #transfer everything from the page start to the page checksum
        #from temp to the final stream
        self.temp.split(self.writer, self.temp,
                        format_size("4b 8u 8u 64S 32u 32u") / 8)

        #write the calculated page checksum to the final stream
        self.writer.write(32, int(self.checksum))

        #transfer everything past the page checksum from temp to final
        self.temp.copy(self.writer)

        #increment sequence number
        self.sequence_number += 1
